@@ -4,7 +4,7 @@
  * This file is part of Libgcrypt.
  *
  * Libgcrypt is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Lesser general Public License as
+ * it under the terms of the GNU Lesser General Public License as
  * published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
@@ -22,16 +22,4 @@
 
 const char *_gcry_compat_identification (void);
 
-
-#ifndef HAVE_GETPID
-pid_t _gcry_getpid (void);
-#define getpid() _gcry_getpid ()
-#endif
-
-#ifndef HAVE_CLOCK
-clock_t _gcry_clock (void);
-#define clock() _gcry_clock ()
-#endif
-
-
 #endif /*GCRY_LIBCOMPAT_H*/
