@@ -32,11 +32,16 @@
 
 static int verbosity_level = 0;
 
+#ifndef HAVE_EXPLICIT_MEMSET
+/* Prevent compiler from optimizing away the call to memset by accessing
+   memset through volatile pointer. */
+static void *(*volatile memset_ptr)(void *, int, size_t) = (void *)memset;
+#endif
+
 static void (*fatal_error_handler)(void*,int, const char*) = NULL;
 static void *fatal_error_handler_value = 0;
 static void (*log_handler)(void*,int, const char*, va_list) = NULL;
 static void *log_handler_value = 0;
-
 static const char *(*user_gettext_handler)( const char * ) = NULL;
 
 void
@@ -91,6 +96,10 @@ _gcry_fatal_error (int rc, const char *text)
   abort ();
 }
 
+/* This is deprecated but very few open source software still uses
+ * this.  However there is more than open source out there and thus we
+ * need to keep the handler specific for Libgcrypt and can't add a
+ * general handler to gpgrt.  */
 void
 _gcry_set_log_handler (void (*f)(void*,int, const char*, va_list), void *opaque)
 {
@@ -110,48 +119,78 @@ _gcry_log_verbosity( int level )
     return verbosity_level >= level;
 }
 
-/****************
- * This is our log function which prints all log messages to stderr or
- * using the function defined with gcry_set_log_handler().
- */
-void
-_gcry_logv( int level, const char *fmt, va_list arg_ptr )
-{
-  if (log_handler)
-    log_handler (log_handler_value, level, fmt, arg_ptr);
-  else
-    {
-      switch (level)
-        {
-        case GCRY_LOG_CONT:  break;
-        case GCRY_LOG_INFO:  break;
-        case GCRY_LOG_WARN:  break;
-        case GCRY_LOG_ERROR: break;
-        case GCRY_LOG_FATAL: fputs("Fatal: ",stderr ); break;
-        case GCRY_LOG_BUG:   fputs("Ohhhh jeeee: ", stderr); break;
-        case GCRY_LOG_DEBUG: fputs("DBG: ", stderr ); break;
-        default: fprintf(stderr,"[Unknown log level %d]: ", level ); break;
-	}
-      vfprintf(stderr,fmt,arg_ptr) ;
-    }
 
-  if ( level == GCRY_LOG_FATAL || level == GCRY_LOG_BUG )
+/* This handler is called after log_fatal and log_info to do what we
+ * used to do in our former own logging functions.  */
+static void
+my_gpgrt_post_fatal_handler (int level)
+{
+  static volatile int done;
+
+  if (!done && (level == GPGRT_LOGLVL_FATAL || level == GPGRT_LOGLVL_BUG))
     {
+      done = 1;
       fips_signal_fatal_error ("internal error (fatal or bug)");
       _gcry_secmem_term ();
-      abort ();
     }
 }
 
 
 void
-_gcry_log( int level, const char *fmt, ... )
+_gcry_set_gpgrt_post_log_handler (void)
 {
-    va_list arg_ptr ;
+  gpgrt_add_post_log_func (my_gpgrt_post_fatal_handler);
+}
 
-    va_start( arg_ptr, fmt ) ;
-    _gcry_logv( level, fmt, arg_ptr );
-    va_end(arg_ptr);
+
+static enum gpgrt_log_levels
+map_log_level (int level)
+{
+  switch (level)
+    {
+    case GCRY_LOG_CONT:  return GPGRT_LOGLVL_CONT;
+    case GCRY_LOG_INFO:  return GPGRT_LOGLVL_INFO;
+    case GCRY_LOG_WARN:  return GPGRT_LOGLVL_WARN;
+    case GCRY_LOG_ERROR: return GPGRT_LOGLVL_ERROR;
+    case GCRY_LOG_FATAL: return GPGRT_LOGLVL_FATAL;
+    case GCRY_LOG_BUG:   return GPGRT_LOGLVL_BUG;
+    case GCRY_LOG_DEBUG:
+    default:             return GPGRT_LOGLVL_DEBUG;
+    }
+}
+
+
+void
+_gcry_logv (int level, const char *fmt, va_list arg_ptr)
+{
+  if (log_handler)
+    {
+      /* The deprecated log handler has been registered.  */
+      log_handler (log_handler_value, level, fmt, arg_ptr);
+      if (level == GCRY_LOG_FATAL || level == GCRY_LOG_BUG)
+        {
+          fips_signal_fatal_error ("internal error (fatal or bug)"
+                                   " [legacy bug handler]");
+          _gcry_secmem_term ();
+          abort ();
+        }
+    }
+  else
+    {
+      gpgrt_logv_domain ("gcrypt", map_log_level (level), NULL, NULL, 0,
+                         fmt, arg_ptr);
+    }
+}
+
+
+void
+_gcry_log (int level, const char *fmt, ...)
+{
+  va_list arg_ptr;
+
+  va_start( arg_ptr, fmt ) ;
+  _gcry_logv (level, fmt, arg_ptr);
+  va_end(arg_ptr);
 }
 
 
@@ -159,7 +198,7 @@ _gcry_log( int level, const char *fmt, ... )
 void
 _gcry_bug( const char *file, int line, const char *func )
 {
-    _gcry_log( GCRY_LOG_BUG,
+    _gcry_log (GCRY_LOG_BUG,
 	     ("... this is a bug (%s:%d:%s)\n"), file, line, func );
     abort(); /* never called, but it makes the compiler happy */
 }
@@ -175,7 +214,7 @@ _gcry_assert_failed (const char *expr, const char *file, int line,
 void
 _gcry_bug( const char *file, int line )
 {
-    _gcry_log( GCRY_LOG_BUG,
+    gpgrt_log( GCRY_LOG_BUG,
 	     _("you found a bug ... (%s:%d)\n"), file, line);
     abort(); /* never called, but it makes the compiler happy */
 }
@@ -194,7 +233,7 @@ _gcry_log_info( const char *fmt, ... )
     va_list arg_ptr ;
 
     va_start( arg_ptr, fmt ) ;
-    _gcry_logv( GCRY_LOG_INFO, fmt, arg_ptr );
+    _gcry_logv (GCRY_LOG_INFO, fmt, arg_ptr);
     va_end(arg_ptr);
 }
 
@@ -204,7 +243,7 @@ _gcry_log_error( const char *fmt, ... )
     va_list arg_ptr ;
 
     va_start( arg_ptr, fmt ) ;
-    _gcry_logv( GCRY_LOG_ERROR, fmt, arg_ptr );
+    _gcry_logv (GCRY_LOG_ERROR, fmt, arg_ptr);
     va_end(arg_ptr);
 }
 
@@ -215,7 +254,7 @@ _gcry_log_fatal( const char *fmt, ... )
     va_list arg_ptr ;
 
     va_start( arg_ptr, fmt ) ;
-    _gcry_logv( GCRY_LOG_FATAL, fmt, arg_ptr );
+    _gcry_logv (GCRY_LOG_FATAL, fmt, arg_ptr);
     va_end(arg_ptr);
     abort(); /* never called, but it makes the compiler happy */
 }
@@ -226,7 +265,7 @@ _gcry_log_bug( const char *fmt, ... )
     va_list arg_ptr ;
 
     va_start( arg_ptr, fmt ) ;
-    _gcry_logv( GCRY_LOG_BUG, fmt, arg_ptr );
+    _gcry_logv (GCRY_LOG_BUG, fmt, arg_ptr);
     va_end(arg_ptr);
     abort(); /* never called, but it makes the compiler happy */
 }
@@ -237,7 +276,7 @@ _gcry_log_debug( const char *fmt, ... )
     va_list arg_ptr ;
 
     va_start( arg_ptr, fmt ) ;
-    _gcry_logv( GCRY_LOG_DEBUG, fmt, arg_ptr );
+    _gcry_logv (GCRY_LOG_DEBUG, fmt, arg_ptr);
     va_end(arg_ptr);
 }
 
@@ -383,8 +422,8 @@ _gcry_log_printsxp (const char *text, gcry_sexp_t sexp)
     {
       int any = 0;
       int n_closing;
-      char *buf, *pend;
-      const char *p;
+      char *buf;
+      const char *p, *pend;
       size_t size;
 
       size = sexp_sprint (sexp, GCRYSEXP_FMT_ADVANCED, NULL, 0);
@@ -498,21 +537,63 @@ _gcry_strtokenize (const char *string, const char *delim)
 
 
 void
+_gcry_fast_wipememory (void *ptr, size_t len)
+{
+  /* Note: This function is called from wipememory/wipememory2 only if LEN
+     is large or unknown at compile time. New wipe function alternatives
+     need to be checked before adding to this function. New implementations
+     need to be faster than wipememory/wipememory2 macros in 'g10lib.h'.
+
+     Following implementations were found to have suboptimal performance:
+
+     - [_WIN32/mingw32] SecureZeroMemory; Inline function, equivalent to
+       volatile byte buffer set: while(buflen--) (volatile char *)(buf++)=set;
+   */
+#ifdef HAVE_EXPLICIT_BZERO
+  explicit_bzero (ptr, len);
+#elif defined(HAVE_EXPLICIT_MEMSET)
+  explicit_memset (ptr, 0, len);
+#else
+  memset_ptr (ptr, 0, len);
+#endif
+}
+
+
+void
+_gcry_fast_wipememory2 (void *ptr, int set, size_t len)
+{
+#ifdef HAVE_EXPLICIT_MEMSET
+  explicit_memset (ptr, set, len);
+#else
+#ifdef HAVE_EXPLICIT_BZERO
+  if (set == 0)
+    {
+      explicit_bzero (ptr, len);
+      return;
+    }
+#endif
+
+  memset_ptr (ptr, set, len);
+#endif
+}
+
+
+void NOINLINE_FUNC
 __gcry_burn_stack (unsigned int bytes)
 {
 #ifdef HAVE_VLA
-    /* (bytes == 0 ? 1 : bytes) == (!bytes + bytes) */
-    unsigned int buflen = ((!bytes + bytes) + 63) & ~63;
-    volatile char buf[buflen];
+  /* (bytes == 0 ? 1 : bytes) == (!bytes + bytes) */
+  unsigned int buflen = ((!bytes + bytes) + 63) & ~63;
+  char buf[buflen];
 
-    wipememory (buf, sizeof buf);
+  _gcry_fast_wipememory (buf, buflen);
 #else
-    volatile char buf[64];
+  volatile char buf[64];
 
-    wipememory (buf, sizeof buf);
+  _gcry_fast_wipememory (buf, sizeof buf);
 
-    if (bytes > sizeof buf)
-        _gcry_burn_stack (bytes - sizeof buf);
+  if (bytes > sizeof buf)
+      _gcry_burn_stack (bytes - sizeof buf);
 #endif
 }
 
